from typing import List, Type
from pydfs_lineup_optimizer.player import LineupPlayer
from pydfs_lineup_optimizer.lineup_printer import BaseLineupPrinter, LineupPrinter


class Lineup(object):
    def __init__(self, players, printer=LineupPrinter):
        # type: (List[LineupPlayer], Type[BaseLineupPrinter]) -> None
        self.players = players
        self.printer = printer()

    def __iter__(self):
        return iter(self.players)

    def __contains__(self, item):
        return item in self.players

    def __str__(self):
        return self.printer.print_lineup(self)

    def __repr__(self):
        return 'Lineup: projection %s, budget %s' % (self.fantasy_points_projection, self.salary_costs)

    @property
    def lineup(self):
        # type: () -> List[LineupPlayer]
        return self.players

    @property
    def fantasy_points_projection(self):
        # type: () -> float
        return round(sum(player.fppg for player in self.players), 3)

    @property
    def salary_costs(self):
        # type: () -> int
        return sum(player.salary for player in self.players)

    def get_unswappable_players(self):
        # type: () -> List[LineupPlayer]
        return [player for player in self.players if player.is_game_started]



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.lineup_printer import IndividualSportLineupPrinter


class DraftKingsSettings(BaseSettings):
    site = Site.DRAFTKINGS
    budget = 50000


@SitesRegistry.register_settings
class DraftKingsBasketballSettings(DraftKingsSettings):
    sport = Sport.BASKETBALL
    positions = [
        LineupPosition('PG', ('PG', )),
        LineupPosition('SG', ('SG', )),
        LineupPosition('SF', ('SF', )),
        LineupPosition('PF', ('PF', )),
        LineupPosition('C', ('C', )),
        LineupPosition('G', ('PG', 'SG')),
        LineupPosition('F', ('SF', 'PF')),
        LineupPosition('UTIL', ('PG', 'SG', 'PF', 'SF', 'C'))
    ]


@SitesRegistry.register_settings
class DraftKingsFootballSettings(DraftKingsSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('QB', ('QB',)),
        LineupPosition('WR1', ('WR',)),
        LineupPosition('WR2', ('WR',)),
        LineupPosition('WR3', ('WR',)),
        LineupPosition('RB1', ('RB',)),
        LineupPosition('RB2', ('RB',)),
        LineupPosition('TE', ('TE',)),
        LineupPosition('FLEX', ('WR', 'RB', 'TE')),
        LineupPosition('DST', ('DST',))
    ]


@SitesRegistry.register_settings
class DraftKingsHockeySettings(DraftKingsSettings):
    sport = Sport.HOCKEY
    positions = [
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('C', ('C',)),
        LineupPosition('C', ('C',)),
        LineupPosition('D', ('D', )),
        LineupPosition('D', ('D', )),
        LineupPosition('G', ('G',)),
        LineupPosition('UTIL', ('LW', 'RW', 'C', 'D'))
    ]


@SitesRegistry.register_settings
class DraftKingsBaseballSettings(DraftKingsSettings):
    sport = Sport.BASEBALL
    max_from_one_team = 5
    positions = [
        LineupPosition('P', ('SP', 'RP')),
        LineupPosition('P', ('SP', 'RP')),
        LineupPosition('C', ('C', )),
        LineupPosition('1B', ('1B', )),
        LineupPosition('2B', ('2B', )),
        LineupPosition('3B', ('3B', )),
        LineupPosition('SS', ('SS', )),
        LineupPosition('OF', ('OF', )),
        LineupPosition('OF', ('OF', )),
        LineupPosition('OF', ('OF', )),
    ]


@SitesRegistry.register_settings
class DraftKingsGolfSettings(DraftKingsSettings):
    sport = Sport.GOLF
    lineup_printer = IndividualSportLineupPrinter
    positions = [
        LineupPosition('G', ('G',)),
        LineupPosition('G', ('G',)),
        LineupPosition('G', ('G',)),
        LineupPosition('G', ('G',)),
        LineupPosition('G', ('G',)),
        LineupPosition('G', ('G',)),
    ]


@SitesRegistry.register_settings
class DraftKingsSoccerSettings(DraftKingsSettings):
    sport = Sport.SOCCER
    positions = [
        LineupPosition('GK', ('GK', )),
        LineupPosition('D', ('D', )),
        LineupPosition('D', ('D', )),
        LineupPosition('M', ('M', )),
        LineupPosition('M', ('M', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('UTIL', ('D', 'M', 'F', )),
    ]


@SitesRegistry.register_settings
class DraftKingsCanadianFootballSettings(DraftKingsSettings):
    sport = Sport.CANADIAN_FOOTBALL
    positions = [
        LineupPosition('QB', ('QB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('FLEX', ('RB', 'WR', )),
        LineupPosition('FLEX', ('RB', 'WR', )),
        LineupPosition('DST', ('DST', )),
    ]


@SitesRegistry.register_settings
class DraftKingsMMA(DraftKingsSettings):
    sport = Sport.MMA
    lineup_printer = IndividualSportLineupPrinter
    positions = [
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
    ]



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.lineup_printer import DropLowestLineupPrinter


class FanDuelSettings(BaseSettings):
    site = Site.FANDUEL
    budget = 60000
    max_from_one_team = 4


@SitesRegistry.register_settings
class FanDuelBasketballSettings(FanDuelSettings):
    sport = Sport.BASKETBALL
    lineup_printer = DropLowestLineupPrinter
    positions = [
        LineupPosition('PG', ('PG', )),
        LineupPosition('PG', ('PG', )),
        LineupPosition('SG', ('SG', )),
        LineupPosition('SG', ('SG', )),
        LineupPosition('SF', ('SF', )),
        LineupPosition('SF', ('SF', )),
        LineupPosition('PF', ('PF', )),
        LineupPosition('PF', ('PF', )),
        LineupPosition('C', ('C', )),
    ]


@SitesRegistry.register_settings
class FanDuelFootballSettings(FanDuelSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('QB', ('QB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('TE', ('TE', )),
        LineupPosition('D', ('D', )),
        LineupPosition('FLEX', ('RB', 'WR', 'TE')),
    ]


@SitesRegistry.register_settings
class FanDuelHockeySettings(FanDuelSettings):
    sport = Sport.HOCKEY
    budget = 55000
    positions = [
        LineupPosition('C', ('C', )),
        LineupPosition('C', ('C', )),
        LineupPosition('W', ('W', )),
        LineupPosition('W', ('W', )),
        LineupPosition('W', ('W', )),
        LineupPosition('W', ('W', )),
        LineupPosition('D', ('D', )),
        LineupPosition('D', ('D', )),
        LineupPosition('G', ('G', )),
    ]


@SitesRegistry.register_settings
class FanDuelBaseballSettings(FanDuelSettings):
    max_from_one_team = 5
    sport = Sport.BASEBALL
    budget = 35000
    positions = [
        LineupPosition('P', ('P',)),
        LineupPosition('C/1B', ('C', '1B')),
        LineupPosition('2B', ('2B',)),
        LineupPosition('3B', ('3B',)),
        LineupPosition('SS', ('SS',)),
        LineupPosition('OF', ('OF',)),
        LineupPosition('OF', ('OF',)),
        LineupPosition('OF', ('OF',)),
        LineupPosition('UTIL', ('1B', '2B', '3B', 'SS', 'C', 'OF')),
    ]


@SitesRegistry.register_settings
class FanDuelWnbaSettings(FanDuelSettings):
    sport = Sport.WNBA
    budget = 40000
    positions = [
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
    ]



from .importer import *
from .settings import *


__all__ = [
    'FantasyDraftCSVImporter', 'FantasyDraftBasketballSettings', 'FantasyDraftFootballSettings',
    'FantasyDraftHockeySettings', 'FantasyDraftBaseballSettings', 'FantasyDraftGolfSettings',
]



from .classic.importer import DraftKingsCSVImporter
from .classic.settings import *
from .captain_mode.importer import DraftKingsCaptainModeCSVImporter
from .captain_mode.settings import *

__all__ = [
    'DraftKingsCSVImporter', 'DraftKingsBasketballSettings', 'DraftKingsFootballSettings',
    'DraftKingsHockeySettings', 'DraftKingsBaseballSettings', 'DraftKingsGolfSettings',
    'DraftKingsSoccerSettings', 'DraftKingsCanadianFootballSettings', 'DraftKingsLOLSettings',
    'DraftKingsCaptainModeCSVImporter', 'DraftKingsCaptainModeFootballSettings',
    'DraftKingsCaptainModeBasketballSettings',
]



from typing import Any
from pydfs_lineup_optimizer.version import __version__
from pydfs_lineup_optimizer.constants import Site, Sport
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException, LineupOptimizerIncorrectTeamName, \
    LineupOptimizerIncorrectPositionName, LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_optimizer import LineupOptimizer
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.sites import SitesRegistry
from pydfs_lineup_optimizer.lineup_exporter import CSVLineupExporter


__all__ = [
    'get_optimizer', 'Site', 'Sport', 'Player', 'LineupOptimizerException', 'LineupOptimizerIncorrectTeamName',
    'LineupOptimizerIncorrectPositionName', 'LineupOptimizerIncorrectCSV', 'LineupOptimizer', 'Lineup',
    'CSVLineupExporter',
]


def get_optimizer(site, sport, **kwargs):
    # type: (str, str, **Any) -> LineupOptimizer
    return LineupOptimizer(SitesRegistry.get_settings(site, sport), **kwargs)



from __future__ import division
from collections import namedtuple
from datetime import datetime
from typing import List, Optional
from pytz import timezone


GameInfo = namedtuple('GameInfo', ['home_team', 'away_team', 'starts_at', 'game_started'])


class Player(object):
    def __init__(self,
                 player_id,  # type: str
                 first_name,  # type: str
                 last_name,  # type: str
                 positions,  # type: List[str]
                 team,  # type: str
                 salary,  # type: float
                 fppg,  # type: float
                 is_injured=False,  # type: bool
                 max_exposure=None,  # type: Optional[float]
                 min_exposure=None,  # type: Optional[float]
                 projected_ownership=None,  # type: Optional[float]
                 game_info=None,  # type: Optional[GameInfo]
                 ):
        # type: (...) -> None
        self.id = player_id
        self.first_name = first_name
        self.last_name = last_name
        self.positions = positions
        self.team = team.upper()
        self.salary = salary
        self.fppg = fppg
        self.is_injured = is_injured
        self.game_info = game_info
        self._max_exposure = None  # type: Optional[float]
        self.max_exposure = max_exposure
        self._min_exposure = None  # type: Optional[float]
        self.min_exposure = min_exposure
        self._projected_ownership = None  # type: Optional[float]
        self.projected_ownership = projected_ownership

    def __repr__(self):
        return '%s %s (%s)' % (self.full_name, '/'.join(self.positions), self.team)

    def __hash__(self):
        return hash(self.id)

    @property
    def max_exposure(self):
        # type: () -> Optional[float]
        return self._max_exposure

    @max_exposure.setter
    def max_exposure(self, max_exposure):
        # type: (Optional[float]) -> None
        self._max_exposure = max_exposure / 100 if max_exposure and max_exposure > 1 else max_exposure

    @property
    def min_exposure(self):
        # type: () -> Optional[float]
        return self._min_exposure

    @min_exposure.setter
    def min_exposure(self, min_exposure):
        # type: (Optional[float]) -> None
        self._min_exposure = min_exposure / 100 if min_exposure and min_exposure > 1 else min_exposure

    @property
    def projected_ownership(self):
        # type: () -> Optional[float]
        return self._projected_ownership

    @projected_ownership.setter
    def projected_ownership(self, projected_ownership):
        # type: (Optional[float]) -> None
        self._projected_ownership = projected_ownership / 100 if projected_ownership and projected_ownership > 1 \
            else projected_ownership

    @property
    def full_name(self):
        # type: () -> str
        return '{} {}'.format(self.first_name, self.last_name)

    @property
    def efficiency(self):
        # type: () -> float
        return round(self.fppg / self.salary, 6)

    @property
    def is_game_started(self):
        # type: () -> bool
        if self.game_info:
            if self.game_info.game_started:
                return True
            starts_at = self.game_info.starts_at
            if starts_at and datetime.now(starts_at.tzinfo) > starts_at:
                return True
        return False


class LineupPlayer(object):
    __slots__ = ['_player', 'lineup_position']

    def __init__(self, player, lineup_position):
        # type: (Player, str) -> None
        self._player = player
        self.lineup_position = lineup_position

    def __getattr__(self, attr_name):
        return getattr(self._player, attr_name)

    def __eq__(self, other):
        if isinstance(other, Player):
            return self._player == other
        elif isinstance(other, LineupPlayer):
            return self._player == other._player
        return NotImplemented

    def __hash__(self):
        return hash(self._player)

    def __repr__(self):
        return repr(self._player)



import csv
from itertools import islice
from datetime import datetime
from pytz import timezone
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.player import Player, LineupPlayer, GameInfo
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.constants import Site


@SitesRegistry.register_csv_importer
class DraftKingsCSVImporter(CSVImporter):  # pragma: nocover
    site = Site.DRAFTKINGS
    DEFAULT_TIMEZONE = 'US/Eastern'

    def _parse_game_info(self, row):
        game_info = row.get('Game Info')
        if not game_info:
            return
        if game_info in ('In Progress', 'Final'):
            return GameInfo(  # No game info provided, just mark game as started
                home_team='',
                away_team='',
                starts_at='',
                game_started=True)
        try:
            teams, date = game_info.split(' ', 1)
            away_team, home_team = teams.split('@')
            starts_at = datetime.strptime(date.replace(' ET', ''), '%m/%d/%Y %I:%M%p').\
                replace(tzinfo=timezone(self.DEFAULT_TIMEZONE))
            return GameInfo(
                home_team=home_team,
                away_team=away_team,
                starts_at=starts_at,
                game_started=False
            )
        except ValueError:
            return

    def _row_to_player(self, row):
        try:
            max_exposure = row.get('Max Exposure', '').replace('%', '')
            name = row['Name'].split()
            player = Player(
                row['ID'],
                name[0],
                name[1] if len(name) > 1 else '',
                row['Position'].split('/'),
                row['TeamAbbrev'],
                float(row['Salary']),
                float(row['AvgPointsPerGame']),
                max_exposure=float(max_exposure) if max_exposure else None,
                game_info=self._parse_game_info(row),
            )
        except KeyError:
            raise LineupOptimizerIncorrectCSV
        return player

    def import_players(self):
        with open(self.filename, 'r') as csv_file:
            start_line = 0  # Find line with 'TeamAbbrev', that's where players data starts
            while True:
                line = csv_file.readline()
                if 'TeamAbbrev' in line:
                    csv_file.seek(0)
                    csv_data = csv.DictReader(islice(csv_file, start_line, None),
                                              skipinitialspace=True)
                    return [self._row_to_player(row) for row in csv_data]
                elif line == '':
                    raise LineupOptimizerIncorrectCSV
                else:
                    start_line += 1

    def import_lineups(self, players):
        with open(self.filename, 'r') as csv_file:
            lines = csv.reader(csv_file)
            try:
                header = next(lines)
                start_column = 4  # First 4 columns has info about tournament
                end_column = header.index('Instructions') - 1
            except (IndexError, ValueError):
                raise LineupOptimizerIncorrectCSV
            position_names = header[start_column:end_column]
            players_dict = {player.id: player for player in players}
            lineups = []
            for line in lines:
                if not line[0]:
                    break
                lineup_players = []
                for index, position in zip(range(start_column, end_column), position_names):
                    try:
                        player_data = line[index]
                        player_data = player_data.replace('(LOCKED)', '')  # Remove possible '(LOCKED)' substring
                        player_id = player_data.split('(')[1][:-1]
                    except IndexError:
                        raise LineupOptimizerIncorrectCSV
                    try:
                        player = players_dict[player_id]
                    except KeyError:
                        raise LineupOptimizerIncorrectCSV('Player not found in players pool')
                    lineup_players.append(LineupPlayer(player, position))
                lineups.append(Lineup(lineup_players))
            return lineups



import csv
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.constants import Site
from pydfs_lineup_optimizer.sites.draftkings.classic.importer import DraftKingsCSVImporter


@SitesRegistry.register_csv_importer
class DraftKingsCaptainModeCSVImporter(DraftKingsCSVImporter):  # pragma: nocover
    site = Site.DRAFTKINGS_CAPTAIN_MODE

    def _row_to_player(self, row):
        try:
            fppg_multiplier = 1.5 if row['Roster Position'] == 'CPT' else 1
            max_exposure = row.get('Max Exposure', '').replace('%', '')
            name = row['Name'].split()
            player = Player(
                row['ID'],
                name[0],
                name[1] if len(name) > 1 else '',
                row['Roster Position'].split('/'),
                row['TeamAbbrev'],
                float(row['Salary']),
                float(row['AvgPointsPerGame']) * fppg_multiplier,
                max_exposure=float(max_exposure) if max_exposure else None
            )
        except KeyError:
            raise LineupOptimizerIncorrectCSV
        return player



from .importer import *
from .settings import *


__all__ = [
    'YahooCSVImporter', 'YahooBasketballSettings', 'YahooFootballSettings', 'YahooHockeySettings',
    'YahooBaseballSettings', 'YahooGolfSettings', 'YahooSoccerSettings',
]



import csv
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.player import Player, GameInfo
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.constants import Site


@SitesRegistry.register_csv_importer
class YahooCSVImporter(CSVImporter):  # pragma: nocover
    site = Site.YAHOO

    def import_players(self):
        players = []
        with open(self.filename, 'r') as csvfile:
            csv_data = csv.DictReader(csvfile, skipinitialspace=True)
            for row in csv_data:
                game_info = None
                try:
                    away_team, home_team = row.get('Game', '').split('@')
                    game_info = GameInfo(home_team, away_team, None, False)
                except ValueError:
                    pass
                try:
                    max_exposure = row.get('Max Exposure')
                    player = Player(
                        row['Id'],
                        row['First Name'],
                        row['Last Name'],
                        row['Position'].split('/'),
                        row['Team'],
                        float(row['Salary']),
                        float(row['FPPG']),
                        True if row['Injury Status'].strip() else False,
                        max_exposure=float(max_exposure.replace('%', '')) if max_exposure else None,
                        game_info=game_info,
                    )
                except KeyError:
                    raise LineupOptimizerIncorrectCSV
                players.append(player)
        return players



from typing import List
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.lineup import Lineup


class CSVImporter(object):
    site = None  # type: str

    def __init__(self, filename):
        # type: (str) -> None
        self.filename = filename

    def import_players(self):  # pragma: no cover
        # type: () -> List[Player]
        raise NotImplementedError

    def import_lineups(self, players):  # pragma: no cover
        # type: (List[Player]) -> List[Lineup]
        raise NotImplementedError



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.lineup_printer import IndividualSportLineupPrinter
from typing import Optional


class FantasyDraftSettings(BaseSettings):
    site = Site.FANTASY_DRAFT
    budget = 100000
    max_from_one_team = 6  # type: Optional[int]


@SitesRegistry.register_settings
class FantasyDraftBasketballSettings(FantasyDraftSettings):
    sport = Sport.BASKETBALL
    positions = [
        LineupPosition('G', ('PG', 'SG')),
        LineupPosition('G', ('PG', 'SG')),
        LineupPosition('G', ('PG', 'SG')),
        LineupPosition('F/C', ('SF', 'PF', 'C')),
        LineupPosition('F/C', ('SF', 'PF', 'C')),
        LineupPosition('F/C', ('SF', 'PF', 'C')),
        LineupPosition('UTIL', ('PG', 'SG', 'SF', 'PF', 'C')),
        LineupPosition('UTIL', ('PG', 'SG', 'SF', 'PF', 'C')),
    ]


@SitesRegistry.register_settings
class FantasyDraftFootballSettings(FantasyDraftSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('QB', ('QB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('TE', ('TE', )),
        LineupPosition('FLEX', ('RB', 'WR', 'TE')),
        LineupPosition('FLEX', ('RB', 'WR', 'TE')),
        LineupPosition('DST', ('DST', ))
    ]


@SitesRegistry.register_settings
class FantasyDraftHockeySettings(FantasyDraftSettings):
    sport = Sport.HOCKEY
    positions = [
        LineupPosition('C', ('C', )),
        LineupPosition('C', ('C', )),
        LineupPosition('W', ('W', )),
        LineupPosition('W', ('W', )),
        LineupPosition('D', ('D', )),
        LineupPosition('UTIL', ('C', 'W', 'D')),
        LineupPosition('UTIL', ('C', 'W', 'D')),
        LineupPosition('Team G', ('TG', )),
    ]


@SitesRegistry.register_settings
class FantasyDraftBaseballSettings(FantasyDraftSettings):
    sport = Sport.BASEBALL
    positions = [
        LineupPosition('P', ('P', )),
        LineupPosition('P', ('P', )),
        LineupPosition('IF', ('1B', '2B', '3B', 'SS', 'IF', 'C',)),
        LineupPosition('IF', ('1B', '2B', '3B', 'SS', 'IF', 'C',)),
        LineupPosition('IF', ('1B', '2B', '3B', 'SS', 'IF', 'C',)),
        LineupPosition('OF', ('LF', 'CF', 'RF', 'OF')),
        LineupPosition('OF', ('LF', 'CF', 'RF', 'OF')),
        LineupPosition('OF', ('LF', 'CF', 'RF', 'OF')),
        LineupPosition('UTIL', ('1B', '2B', '3B', 'SS', 'IF', 'C', 'LF', 'CF', 'RF', 'OF')),
        LineupPosition('UTIL', ('1B', '2B', '3B', 'SS', 'IF', 'C', 'LF', 'CF', 'RF', 'OF')),
    ]


@SitesRegistry.register_settings
class FantasyDraftGolfSettings(FantasyDraftSettings):
    sport = Sport.GOLF
    max_from_one_team = None
    lineup_printer = IndividualSportLineupPrinter
    positions = [
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
    ]



from collections import defaultdict
from typing import Type, DefaultDict, Dict
from pydfs_lineup_optimizer.settings import BaseSettings
from pydfs_lineup_optimizer.lineup_importer import CSVImporter


class SitesRegistry(object):
    SETTINGS_MAPPING = defaultdict(dict)  # type: DefaultDict[str, Dict[str, Type[BaseSettings]]]
    CSV_IMPORTERS_MAPPING = {}  # type: Dict[str, Type[CSVImporter]]

    @classmethod
    def register_settings(cls, settings_cls):
        # type: (Type[BaseSettings]) -> Type[BaseSettings]
        cls.SETTINGS_MAPPING[settings_cls.site][settings_cls.sport] = settings_cls
        return settings_cls

    @classmethod
    def register_csv_importer(cls, importer_cls):
        # type: (Type[CSVImporter]) -> Type[CSVImporter]
        cls.CSV_IMPORTERS_MAPPING[importer_cls.site] = importer_cls
        return importer_cls

    @classmethod
    def get_settings(cls, site, sport):
        # type: (str, str) -> Type[BaseSettings]
        try:
            return cls.SETTINGS_MAPPING[site][sport]
        except KeyError:
            raise NotImplementedError

    @classmethod
    def get_csv_importer(cls, site):
        # type: (str) -> Type[CSVImporter]
        try:
            return cls.CSV_IMPORTERS_MAPPING[site]
        except KeyError:
            raise NotImplementedError



from typing import TypeVar, Any, List, Iterable, Optional


Self = TypeVar('Self', bound='Solver')


class Solver(object):  # pragma: no cover
    def setup_solver(self):
        # type: () -> None
        raise NotImplementedError

    def set_objective(self, variables, coefficients):
        # type: (Iterable[Any], Iterable[float]) -> None
        raise NotImplementedError

    def add_variable(self, name):
        # type: (str) -> Any
        raise NotImplementedError

    def add_constraint(self, variables, coefficients, sense, rhs):
        # type: (Iterable[Any], Optional[Iterable[float]], str, float) -> None
        raise NotImplementedError

    def solve(self):
        # type: () -> List[Any]
        raise NotImplementedError

    def copy(self):
        # type: () -> Self
        raise NotImplementedError



import csv
from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.constants import Site


@SitesRegistry.register_csv_importer
class FantasyDraftCSVImporter(CSVImporter):  # pragma: nocover
    site = Site.FANTASY_DRAFT

    def import_players(self):
        players = []
        with open(self.filename, 'r') as csvfile:
            csv_data = csv.DictReader(csvfile, skipinitialspace=True)
            for i, row in enumerate(csv_data):
                try:
                    name = row['Name'].split()
                    max_exposure = row.get('Max Exposure')
                    fppg = row.get('Avg FPPG') or row.get('Avg FPPT')
                    if not fppg:
                        raise LineupOptimizerIncorrectCSV
                    player = Player(
                        str(i),
                        name[0],
                        name[1] if len(name) > 1 else '',
                        row['Position'].split('/'),
                        row.get('Team', row.get('Game', '')),
                        float(row['Salary'].replace('$', '').replace(',', '')),
                        float(fppg),
                        max_exposure=float(max_exposure.replace('%', '')) if max_exposure else None
                    )
                except KeyError:
                    raise LineupOptimizerIncorrectCSV
                players.append(player)
        return players



from .importer import *
from .settings import *


__all__ = [
    'FanDuelCSVImporter', 'FanDuelBasketballSettings', 'FanDuelFootballSettings', 'FanDuelHockeySettings',
    'FanDuelBaseballSettings', 'FanDuelWnbaSettings',
]



import multiprocessing
from pulp import LpProblem, LpMaximize, LpVariable, lpSum, LpStatusOptimal, LpBinary
from pulp.solvers import PULP_CBC_CMD
from .base import Solver
from .constants import SolverSign
from .exceptions import SolverException


class PuLPSolver(Solver):
    LP_SOLVER = PULP_CBC_CMD(threads=multiprocessing.cpu_count())

    def __init__(self):
        self.prob = None

    def setup_solver(self):
        self.prob = LpProblem('Daily Fantasy Sports', LpMaximize)

    def add_variable(self, name):
        return LpVariable(name, cat=LpBinary)

    def set_objective(self, variables, coefficients):
        self.prob += lpSum([variable * coefficient for variable, coefficient in zip(variables, coefficients)])

    def add_constraint(self, variables, coefficients, sign, rhs):
        if coefficients:
            lhs = [variable * coefficient for variable, coefficient in zip(variables, coefficients)]
        else:
            lhs = variables
        if sign == SolverSign.EQ:
            self.prob += lpSum(lhs) == rhs
        elif sign == SolverSign.NOT_EQ:
            self.prob += lpSum(lhs) != rhs
        elif sign == SolverSign.GTE:
            self.prob += lpSum(lhs) >= rhs
        elif sign == SolverSign.LTE:
            self.prob += lpSum(lhs) <= rhs
        else:
            raise SolverException('Incorrect constraint sign')

    def copy(self):
        new_solver = type(self)()
        new_solver.prob = self.prob.copy()
        return new_solver

    def solve(self):
        self.prob.solve(self.LP_SOLVER)
        if self.prob.status == LpStatusOptimal:
            result = []
            for variable in self.prob.variables():
                if variable.value() == 1.0:
                    result.append(variable)
            return result
        else:
            raise SolverException('Unable to solve')



from __future__ import division
from collections import Counter, OrderedDict
from itertools import chain
from typing import FrozenSet, Tuple, Type, Generator, Set
from pydfs_lineup_optimizer.solvers import PuLPSolver, SolverException
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException, LineupOptimizerIncorrectTeamName, \
    LineupOptimizerIncorrectPositionName
from pydfs_lineup_optimizer.sites import SitesRegistry
from pydfs_lineup_optimizer.lineup_importer import CSVImporter
from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.player import LineupPlayer, GameInfo
from pydfs_lineup_optimizer.utils import ratio, link_players_with_positions, get_remaining_positions
from pydfs_lineup_optimizer.rules import *


BASE_CONSTRAINTS = {TotalPlayersRule, LineupBudgetRule, PositionsRule, MaxFromOneTeamRule, LockedPlayersRule,
                    RemoveInjuredRule, UniquePlayerRule}


class LineupOptimizer(object):
    def __init__(self, settings, solver=PuLPSolver):
        # type: (Type[BaseSettings], Type[Solver]) -> None
        self._settings = settings
        self._csv_importer = None  # type: Optional[Type[CSVImporter]]
        self._constraints = BASE_CONSTRAINTS.copy()  # type: Set[Type[OptimizerRule]]
        self._players = []  # type: List[Player]
        self._lineup = []  # type: List[Player]
        self._available_positions = frozenset(chain.from_iterable(
            position.positions for position in self._settings.positions))
        self._removed_players = []  # type: List[Player]
        self._search_threshold = 0.8
        self._min_deviation = 0.06
        self._max_deviation = 0.12
        self._players_from_one_team = {}  # type: Dict[str, int]
        self._players_with_same_position = {}  # type: Dict[str, int]
        self._positions_from_same_team = []  # type: List[str]
        self._min_salary_cap = None  # type: Optional[float]
        self._max_repeating_players = None  # type: Optional[int]
        self._solver_class = solver
        self._max_projected_ownership = None  # type: Optional[float]
        self._min_projected_ownership = None  # type: Optional[float]
        self._team_stacks = None  # type: Optional[List[int]]
        self._opposing_teams_position_restriction = None  # type: Optional[Tuple[List[str], List[str]]]

    @property
    def budget(self):
        # type: () -> float
        return self._settings.budget

    @property
    def total_players(self):
        # type: () -> int
        return self._settings.get_total_players()

    @property
    def remaining_budget(self):
        # type: () -> float
        return self.budget - sum(player.salary for player in self.locked_players)

    @property
    def remaining_players(self):
        # type: () -> int
        return self.total_players - len(self.locked_players)

    @property
    def max_from_one_team(self):
        # type: () -> Optional[int]
        return self._settings.max_from_one_team

    @property
    def positions_from_same_team(self):
        # type: () -> List[str]
        return self._positions_from_same_team

    @property
    def players_from_one_team(self):
        # type: () -> Dict[str, int]
        return self._players_from_one_team

    @property
    def players_with_same_position(self):
        # type: () -> Dict[str, int]
        return self._players_with_same_position

    @property
    def available_teams(self):
        # type: () -> FrozenSet[str]
        return frozenset([p.team for p in self._players])

    @property
    def available_positions(self):
        # type: () -> FrozenSet[LineupPosition]
        return self._available_positions

    @property
    def removed_players(self):
        # type: () -> List[Player]
        return self._removed_players

    @property
    def players(self):
        # type: () -> List[Player]
        return [player for player in self._players if player not in self.removed_players]

    @property
    def locked_players(self):
        # type: () -> List[Player]
        return self._lineup

    @property
    def min_salary_cap(self):
        # type: () -> Optional[float]
        return self._min_salary_cap

    @property
    def max_repeating_players(self):
        # type: () -> Optional[int]
        return self._max_repeating_players

    @property
    def max_projected_ownership(self):
        # type: () -> Optional[float]
        return self._max_projected_ownership

    @property
    def min_projected_ownership(self):
        # type: () -> Optional[float]
        return self._min_projected_ownership

    @property
    def settings(self):
        # type: () -> Type[BaseSettings]
        return self._settings

    @property
    def team_stacks(self):
        # type: () -> Optional[List[int]]
        return self._team_stacks

    @property
    def opposing_teams_position_restriction(self):
        # type: () -> Optional[Tuple[List[str], List[str]]]
        return self._opposing_teams_position_restriction

    @property
    def games(self):
        # type: () -> FrozenSet[GameInfo]
        return frozenset(player.game_info for player in self.players if player.game_info)

    def reset_lineup(self):
        self._lineup = []
        self._players_with_same_position = {}
        self._players_from_one_team = {}
        self._positions_from_same_team = []

    def set_deviation(self, min_deviation, max_deviation):
        # type: (float, float) -> None
        """
        Set deviation ranges for randomness mode
        """
        self._min_deviation = min_deviation
        self._max_deviation = max_deviation

    def set_csv_importer(self, csv_importer):
        # type: (Type[CSVImporter]) -> None
        self._csv_importer = csv_importer

    def get_deviation(self):
        # type: () -> Tuple[float, float]
        return self._min_deviation, self._max_deviation

    def set_min_salary_cap(self, min_salary):
        # type: (float) -> None
        if min_salary > self.budget:
            raise LineupOptimizerException('Min salary greater than max budget')
        self.add_new_rule(MinSalaryCapRule)
        self._min_salary_cap = min_salary

    def load_players_from_csv(self, filename):
        # type: (str) -> None
        """
        Load player list from CSV file with passed filename.
        """
        csv_importer = self._csv_importer
        if not csv_importer:
            csv_importer = SitesRegistry.get_csv_importer(self._settings.site)
        self._players = csv_importer(filename).import_players()

    def load_lineups_from_csv(self, filename):
        # type: (str) -> List[Lineup]
        csv_importer = self._csv_importer
        if not csv_importer:
            csv_importer = SitesRegistry.get_csv_importer(self._settings.site)
        return csv_importer(filename).import_lineups(self.players)

    def load_players(self, players):
        # type: (List[Player]) -> None
        """
        Manually loads player to optimizer
        """
        self._players = players

    def extend_players(self, players):
        # type: (List[Player]) -> None
        """
        Add more players for current optimizer players
        """
        self._players.extend(players)

    def add_new_rule(self, rule):
        # type: (Type[OptimizerRule]) -> None
        self._constraints.add(rule)

    def remove_rule(self, rule, silent=True):
        # type: (Type[OptimizerRule], bool) -> None
        try:
            self._constraints.remove(rule)
        except KeyError:
            if not silent:
                raise LineupOptimizerException('Rule isn\'t added!')

    def remove_player(self, player):
        # type: (Player) -> None
        """
        Remove player from list for selecting players for lineup.
        """
        self._removed_players.append(player)

    def restore_player(self, player):
        # type: (Player) -> None
        """
        Restore removed player.
        """
        try:
            self._removed_players.remove(player)
        except ValueError:
            raise LineupOptimizerException('Player not removed!')

    def find_players(self, name):
        # type: (str) -> List[Player]
        """
        Return list of players with similar name.
        """
        possibilities = [(player, ratio(name, player.full_name)) for player in self._players]
        filtered_possibilities = filter(lambda pos: pos[1] >= self._search_threshold, possibilities)
        players = sorted(filtered_possibilities, key=lambda pos: -pos[1])
        return list(map(lambda p: p[0], players))

    def get_player_by_name(self, name):
        # type: (str) -> Optional[Player]
        """
        Return closest player with similar name or None.
        """
        players = self.find_players(name)
        return players[0] if players else None

    def add_player_to_lineup(self, player):
        # type: (Player) -> None
        """
        Forces adding specified player to lineup.
        Return true if player successfully added to lineup.
        """
        if player.max_exposure == 0:
            raise LineupOptimizerException('Can\'t add this player to line up! Player has max exposure set to 0.')
        if player in self._lineup:
            raise LineupOptimizerException('This player already in your line up!')
        if player.salary > self.remaining_budget:
            raise LineupOptimizerException('Can\'t add this player to line up! Your team is over budget!')
        if self.remaining_players < 1:
            raise LineupOptimizerException('Can\'t add this player to line up! You already select all %s players!' %
                                           len(self.locked_players))
        if self.max_from_one_team:
            from_same_team = len([p for p in self.locked_players if p.team == player.team])
            if from_same_team + 1 > self.max_from_one_team:
                raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                               self.max_from_one_team)
        try:
            link_players_with_positions(self.locked_players + [player], self._settings.positions)
        except LineupOptimizerException:
            raise LineupOptimizerException('You\'re already select all %s\'s' % '/'.join(player.positions))
        self._lineup.append(player)

    def remove_player_from_lineup(self, player):
        # type: (Player) -> None
        """
        Removes specified player from lineup.
        """
        try:
            self._lineup.remove(player)
        except ValueError:
            raise LineupOptimizerException('Player not in line up!')

    def set_players_from_one_team(self, teams=None):
        # type: (Optional[Dict[str, int]]) -> None
        if teams is not None:
            teams = {team.upper(): num_of_players for team, num_of_players in teams.items()}
            for team, num_of_players in teams.items():
                self._check_team_constraint(team, num_of_players)
            self.add_new_rule(TeamMatesRule)
        else:
            self.remove_rule(TeamMatesRule)
        self._players_from_one_team = teams or {}

    def set_players_with_same_position(self, positions):
        # type: (Dict[str, int]) -> None
        positions = positions or {}
        positions = {position.upper(): num_of_players for position, num_of_players in positions.items()}
        for pos, val in positions.items():
            self._check_position_constraint(pos)
        self._players_with_same_position = positions

    def set_positions_for_same_team(self, positions):
        # type: (Optional[List[str]]) -> None
        if positions is not None:
            if self.max_from_one_team and len(positions) > self.max_from_one_team:
                raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                               self.max_from_one_team)
            positions = [position.upper() for position in positions]
            for position, num_of_players in Counter(positions).items():
                self._check_position_constraint(position)
            self.add_new_rule(FromSameTeamByPositionsRule)
        else:
            self.remove_rule(FromSameTeamByPositionsRule)
        self._positions_from_same_team = positions or []

    def set_max_repeating_players(self, max_repeating_players):
        # type: (int) -> None
        if max_repeating_players >= self.total_players:
            raise LineupOptimizerException('Maximum repeating players should be smaller than %d' % self.total_players)
        elif max_repeating_players < 1:
            raise LineupOptimizerException('Maximum repeating players should be 1 or greater')
        self._max_repeating_players = max_repeating_players
        self.add_new_rule(MaxRepeatingPlayersRule)

    def set_projected_ownership(self, min_projected_ownership=None, max_projected_ownership=None):
        # type: (Optional[float], Optional[float]) -> None
        if min_projected_ownership and max_projected_ownership and min_projected_ownership >= max_projected_ownership:
            raise LineupOptimizerException('Max projected ownership should be greater than min projected ownership')
        self._max_projected_ownership = max_projected_ownership / 100 if \
            max_projected_ownership and max_projected_ownership > 1 else max_projected_ownership
        self._min_projected_ownership = min_projected_ownership / 100 if \
            min_projected_ownership and min_projected_ownership > 1 else min_projected_ownership
        if max_projected_ownership or min_projected_ownership:
            self.add_new_rule(ProjectedOwnershipRule)
        else:
            self.remove_rule(ProjectedOwnershipRule)

    def set_team_stacking(self, stacks):
        # type: (Optional[List[int]]) -> None
        if stacks:
            if sum(stacks) > self.settings.get_total_players():
                raise LineupOptimizerException('Sum of stacks should be less than %d' % self.total_players)
            max_from_one_team = self.settings.max_from_one_team
            if max_from_one_team and any([stack > max_from_one_team for stack in stacks]):
                raise LineupOptimizerException('Stacks should be smaller than max players from one team (%d)' %
                                               self.total_players)
            self.add_new_rule(TeamStacksRule)
        else:
            self.remove_rule(TeamStacksRule)
        self._team_stacks = stacks

    def restrict_positions_for_opposing_team(self, first_team_positions, second_team_positions):
        # type: (List[str], List[str]) -> None
        if not self.games:
            raise LineupOptimizerException('Game Info isn\'t specified for players')
        self._opposing_teams_position_restriction = (first_team_positions, second_team_positions)
        self.add_new_rule(RestrictPositionsForOpposingTeams)

    def optimize(self, n, max_exposure=None, randomness=False, with_injured=False):
        # type: (int, Optional[float], bool, bool) -> Generator[Lineup, None, None]
        params = locals()
        rules = self._constraints.copy()
        if randomness:
            rules.add(RandomObjective)
        else:
            rules.add(NormalObjective)
        if with_injured:
            rules.remove(RemoveInjuredRule)
        players = [player for player in self.players if player.max_exposure is None or player.max_exposure > 0]
        base_solver = self._solver_class()
        base_solver.setup_solver()
        players_dict = OrderedDict(
            [(player, base_solver.add_variable('Player_%d' % i)) for i, player in enumerate(players)])
        variables_dict = {v: k for k, v in players_dict.items()}
        constraints = [constraint(self, params) for constraint in rules]
        for constraint in constraints:
            constraint.apply(base_solver, players_dict)
        previous_lineup = None
        for _ in range(n):
            solver = base_solver.copy()  # type: Solver
            for constraint in constraints:
                constraint.apply_for_iteration(solver, players_dict, previous_lineup)
            try:
                solved_variables = solver.solve()
                lineup_players = []
                for solved_variable in solved_variables:
                    player = variables_dict.get(solved_variable)
                    if player:
                        lineup_players.append(player)
                lineup = self._build_lineup(lineup_players)
                previous_lineup = lineup
                yield lineup
                if len(self.locked_players) == self.total_players:
                    return
            except SolverException:
                raise LineupOptimizerException('Can\'t generate lineups')

    def optimize_lineups(self, lineups):
        params = {
            'n': len(lineups),
            'lineups': lineups,
        }
        rules = self._constraints.copy()
        rules.add(NormalObjective)
        rules.add(LateSwapRule)
        rules.remove(PositionsRule)
        players = [player for player in self.players if player.max_exposure is None or player.max_exposure > 0]
        base_solver = self._solver_class()
        base_solver.setup_solver()
        players_dict = OrderedDict(
            [(player, base_solver.add_variable('Player_%d' % i)) for i, player in enumerate(players)])
        variables_dict = {v: k for k, v in players_dict.items()}
        constraints = [constraint(self, params) for constraint in rules]
        for constraint in constraints:
            constraint.apply(base_solver, players_dict)
        previous_lineup = None
        for lineup in lineups:
            solver = base_solver.copy()
            for constraint in constraints:
                constraint.apply_for_iteration(solver, players_dict, previous_lineup)
            try:
                solved_variables = solver.solve()
                unswappable_players = lineup.get_unswappable_players()
                lineup_players = []
                for solved_variable in solved_variables:
                    player = variables_dict.get(solved_variable)
                    if player:
                        lineup_players.append(player)
                generated_lineup = self._build_lineup(lineup_players, unswappable_players)
                previous_lineup = generated_lineup
                yield generated_lineup
                if len(self.locked_players) == self.total_players:
                    return
            except SolverException:
                raise LineupOptimizerException('Can\'t generate lineups')

    def _build_lineup(self, players, unswappable_players=None):
        # type: (List[Player], Optional[List[LineupPlayer]]) -> Lineup
        lineup = []
        positions = self._settings.positions[:]
        if unswappable_players:
            players = [player for player in players if player not in unswappable_players]
            positions = get_remaining_positions(positions, unswappable_players)
            lineup.extend(unswappable_players)
        players_with_positions = link_players_with_positions(players, positions)
        for player, position in players_with_positions.items():
            lineup.append(LineupPlayer(player, position.name))
        positions_order = [pos.name for pos in self._settings.positions]
        lineup.sort(key=lambda p: positions_order.index(p.lineup_position))
        return Lineup(lineup, self._settings.lineup_printer)

    def _check_team_constraint(self, team, num_of_players):
        # type: (str, int) -> None
        if team not in self.available_teams:
            raise LineupOptimizerIncorrectTeamName('%s is incorrect team name.' % team)
        if self.max_from_one_team and num_of_players > self.max_from_one_team:
            raise LineupOptimizerException('You can\'t set more than %s players from one team.' %
                                           self.max_from_one_team)

    def _check_position_constraint(self, position):
        # type: (str) -> None
        if position not in self.available_positions:
            raise LineupOptimizerIncorrectPositionName('%s is incorrect position name.' % position)



import csv
from typing import Iterable, Callable
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.player import LineupPlayer


class LineupExporter(object):
    def __init__(self, lineups):
        # type: (Iterable[Lineup]) -> None
        self.lineups = lineups

    @staticmethod
    def render_player(player):
        # type: (LineupPlayer) -> str
        result = player.full_name  # type: str
        if player.id:
            result += '(%s)' % player.id
        return result

    def export(self, filename, render_func=None):
        # type: (str, Callable[[LineupPlayer], str]) -> None
        raise NotImplementedError


class CSVLineupExporter(LineupExporter):
    def export(self, filename, render_func=None):
        # type: (str, Callable[[LineupPlayer], str]) -> None
        with open(filename, 'w') as csvfile:
            lineup_writer = csv.writer(csvfile, delimiter=',')
            for index, lineup in enumerate(self.lineups):
                if index == 0:
                    header = [player.lineup_position for player in lineup.lineup]
                    header.extend(('Budget', 'FPPG'))
                    lineup_writer.writerow(header)
                row = [(render_func or self.render_player)(player) for player in lineup.lineup]
                row.append(str(lineup.salary_costs))
                row.append(str(lineup.fantasy_points_projection))
                lineup_writer.writerow(row)



from typing import Dict, Tuple, List, Iterable, Any
from collections import Counter
from difflib import SequenceMatcher
from itertools import combinations, chain, permutations
from pydfs_lineup_optimizer.player import Player, LineupPlayer
from pydfs_lineup_optimizer.settings import LineupPosition
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException


def list_intersection(first_list, second_list):
    # type: (Iterable[Any], Iterable[Any]) -> bool
    for el in first_list:
        if el in second_list:
            return True
    return False


def ratio(search_string, possible_match):
    # type: (str, str) -> float
    search_string = search_string.lower()
    possible_match = possible_match.lower()
    if len(search_string) >= len(possible_match):
        parts = [possible_match]
    else:
        shorter_length = len(search_string)
        num_of_parts = len(possible_match) - shorter_length
        parts = [possible_match[i:i + shorter_length] for i in range(num_of_parts + 1)]
    return max([SequenceMatcher(None, search_string, part).ratio() for part in parts])


def get_positions_for_optimizer(positions_list):
    # type: (List[LineupPosition]) -> Dict[Tuple[str, ...], int]
    """
    Convert positions list into dict for using in optimizer.
    """
    positions = {}
    positions_counter = Counter([tuple(sorted(p.positions)) for p in positions_list])
    for key in positions_counter.keys():
        min_value = positions_counter[key] + len(list(filter(
            lambda p: len(p.positions) < len(key) and list_intersection(key, p.positions), positions_list
        )))
        positions[key] = min_value
    for i in range(2, len(positions)):
        for positions_tuple in combinations(positions_counter.keys(), i):
            flatten_positions = tuple(sorted(chain.from_iterable(positions_tuple)))
            if len(flatten_positions) != len(set(flatten_positions)):
                continue
            if flatten_positions in positions:
                continue
            min_value = sum(positions[pos] for pos in positions_tuple)
            positions[flatten_positions] = min_value
    return positions


def link_players_with_positions(players, positions):
    # type: (List[Player], List[LineupPosition]) -> Dict[Player, LineupPosition]
    """
    This method tries to set positions for given players, and raise error if can't.
    """
    positions = positions[:]
    single_position_players = []  # type: List[Player]
    multi_positions_players = []  # type: List[Player]
    players_with_positions = {}  # type: Dict[Player, LineupPosition]
    for player in players:
        if len(player.positions) == 1:
            single_position_players.append(player)
        else:
            multi_positions_players.append(player)
    for player in single_position_players:
        for position in positions:
            if player.positions[0] in position.positions:
                players_with_positions[player] = position
                positions.remove(position)
                break
        else:
            raise LineupOptimizerException('Unable to build lineup')
    for players_permutation in permutations(multi_positions_players):
        is_correct = True
        remaining_positions = positions[:]
        for player in players_permutation:
            for position in remaining_positions:
                if list_intersection(player.positions, position.positions):
                    players_with_positions[player] = position
                    remaining_positions.remove(position)
                    break
            else:
                is_correct = False
                break
        if is_correct:
            break
    else:
        raise LineupOptimizerException('Unable to build lineup')
    return players_with_positions


def get_remaining_positions(positions, unswappable_players):
    # type: (List[LineupPosition], List[LineupPlayer]) -> List[LineupPosition]
    """
    Remove unswappable players positions from positions list
    """
    positions = positions[:]
    for player in unswappable_players:
        for position in positions:
            if position.name == player.lineup_position:
                positions.remove(position)
                break
    return positions



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry


class FanBallSettings(BaseSettings):
    site = Site.FANBALL
    budget = 55000


@SitesRegistry.register_settings
class FanBallFootballSettings(FanBallSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('QB', ('QB',)),
        LineupPosition('WR1', ('WR',)),
        LineupPosition('WR2', ('WR',)),
        LineupPosition('WR3', ('WR',)),
        LineupPosition('RB1', ('RB',)),
        LineupPosition('RB2', ('RB',)),
        LineupPosition('TE', ('TE',)),
        LineupPosition('FLEX', ('WR', 'RB', 'TE')),
        LineupPosition('DST', ('DST',))
    ]





from .sites_registry import SitesRegistry
from .draftkings import *  # type: ignore
from .fanball import *  # type: ignore
from .fanduel import *  # type: ignore
from .fantasy_draft import *  # type: ignore
from .yahoo import *  # type: ignore



class LineupOptimizerException(Exception):
    def __init__(self, message):
        # type: (str) -> None
        self.message = message

    def __str__(self):
        return self.message


class LineupOptimizerIncorrectTeamName(LineupOptimizerException):
    pass


class LineupOptimizerIncorrectPositionName(LineupOptimizerException):
    pass


class LineupOptimizerIncorrectCSV(LineupOptimizerException):
    def __init__(self, message='Incorrect csv format!'):
        # type: (str) -> None
        super(LineupOptimizerIncorrectCSV, self).__init__(message)



from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.lineup import Lineup, LineupPlayer


class BaseLineupPrinter(object):
    def print_lineup(self, lineup):
        # type: ('Lineup') -> str
        raise NotImplementedError


class LineupPrinter(BaseLineupPrinter):
    OUTPUT_FORMAT = '{0:>2}. {1:<5} {2:<30}{3:<6}{4:<15}{5:<9}{6:<8}{7:<10}\n'

    def _print_game_info(self, player):
        # type: ('LineupPlayer') -> str
        game_info = player.game_info
        if game_info:
            return '%s@%s' % (game_info.away_team, game_info.home_team)
        return ''

    def _print_player(self, index, player):
        # type: (int, 'LineupPlayer') -> str
        return self.OUTPUT_FORMAT.format(
            index,
            player.lineup_position,
            player.full_name,
            '/'.join(player.positions),
            player.team,
            self._print_game_info(player),
            round(player.fppg, 3),
            str(player.salary) + '$',
        )

    def print_lineup(self, lineup):
        res = ''
        for index, player in enumerate(lineup.players, start=1):
            res += self._print_player(index, player)
        res += '\nFantasy Points %.2f' % lineup.fantasy_points_projection
        res += '\nSalary %.2f\n' % lineup.salary_costs
        return res


class DropLowestLineupPrinter(LineupPrinter):
    def _print_player(self, index, player, is_dropped=False):
        return self.OUTPUT_FORMAT.format(
            index,
            player.lineup_position,
            '%s%s' % (player.full_name, '(DROPPED)' if is_dropped else ''),
            '/'.join(player.positions),
            player.team,
            self._print_game_info(player),
            round(player.fppg, 3),
            str(player.salary) + '$',
        )

    def print_lineup(self, lineup):
        res = ''
        lowest_fppg_player = sorted(lineup, key=lambda p: p.fppg)[0]
        for index, player in enumerate(lineup.players, start=1):
            res += self._print_player(index, player, is_dropped=lowest_fppg_player)
        res += 'Fantasy Points %.2f' % lineup.fantasy_points_projection
        res += '\nFantasy Points Without Dropped Player %.2f' % \
               (lineup.fantasy_points_projection - lowest_fppg_player.fppg)
        res += '\nSalary %.2f\n' % lineup.salary_costs
        return res


class IndividualSportLineupPrinter(LineupPrinter):
    def _print_player(self, index, player):
        return '{0:>2}. {1:<5} {2:<30}{3:<8}{4:<10}\n'.format(
            index,
            player.lineup_position,
            player.full_name,
            round(player.fppg, 3),
            str(player.salary) + '$',
        )



import os
from setuptools import setup, find_packages


exec(open('pydfs_lineup_optimizer/version.py').read())


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='pydfs-lineup-optimizer',
    version=__version__,
    packages=find_packages(exclude=['tests*']),
    url='https://github.com/DimaKudosh/pydfs-lineup-optimizer',
    license='MIT',
    author='Dima Kudosh',
    author_email='dimakudosh@gmail.com',
    description='Tool for creating optimal lineups for daily fantasy sports',
    keywords=['dfs', 'fantasy', 'sport', 'lineup', 'optimize', 'optimizer', 'nba', 'nfl', 'nhl', 'mlb'],
    install_requires=['PuLP>=1.6.8', 'typing', 'pytz'],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Programming Language :: Python',
    ],
)



from collections import namedtuple
from typing import Optional, Type, List
from pydfs_lineup_optimizer.lineup_printer import LineupPrinter, BaseLineupPrinter


LineupPosition = namedtuple('LineupPosition', ['name', 'positions'])


class BaseSettings(object):
    site = None  # type: str
    sport = None  # type: str
    budget = 0  # type: float
    positions = []  # type: List[LineupPosition]
    max_from_one_team = None  # type: Optional[int]
    lineup_printer = LineupPrinter  # type: Type[BaseLineupPrinter]

    @classmethod
    def get_total_players(cls):
        # type: () -> int
        return len(cls.positions)



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry


class DraftKingsCaptainModeSettings(BaseSettings):
    site = Site.DRAFTKINGS_CAPTAIN_MODE
    budget = 50000
    max_from_one_team = 5


@SitesRegistry.register_settings
class DraftKingsCaptainModeFootballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('CPT', ('CPT', )),
        LineupPosition('FLEX', ('FLEX', )),
        LineupPosition('FLEX', ('FLEX', )),
        LineupPosition('FLEX', ('FLEX', )),
        LineupPosition('FLEX', ('FLEX', )),
        LineupPosition('FLEX', ('FLEX', )),
    ]


@SitesRegistry.register_settings
class DraftKingsCaptainModeBasketballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.BASKETBALL
    positions = [
        LineupPosition('CPT', ('CPT', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
    ]


@SitesRegistry.register_settings
class DraftKingsLOLSettings(DraftKingsCaptainModeSettings):
    sport = Sport.LEAGUE_OF_LEGENDS
    max_from_one_team = 4
    positions = [
        LineupPosition('CPT', ('CPT', )),
        LineupPosition('TOP', ('TOP', )),
        LineupPosition('JNG', ('JNG', )),
        LineupPosition('MID', ('MID', )),
        LineupPosition('ADC', ('ADC', )),
        LineupPosition('SUP', ('SUP', )),
        LineupPosition('TEAM', ('TEAM', )),
    ]


@SitesRegistry.register_settings
class DraftKingsCaptainModeBaseballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.BASEBALL
    positions = [
        LineupPosition('CPT', ('CPT', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
        LineupPosition('UTIL', ('UTIL', )),
    ]



from __future__ import division
from collections import defaultdict
from itertools import product, combinations, groupby, permutations
from math import ceil
from random import getrandbits, uniform
from typing import List, Dict, DefaultDict, Any, Optional, TYPE_CHECKING
from pydfs_lineup_optimizer.solvers import Solver, SolverSign
from pydfs_lineup_optimizer.utils import list_intersection, get_positions_for_optimizer, get_remaining_positions
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.player import Player


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.lineup_optimizer import LineupOptimizer


class OptimizerRule(object):
    def __init__(self, optimizer, params):
        # type: ('LineupOptimizer', Dict[str, Any]) -> None
        self.params = params
        self.optimizer = optimizer

    def apply(self, solver, players_dict):
        # type: (Solver, Dict[Player, Any]) -> None
        pass

    def apply_for_iteration(self, solver, players_dict, result):
        # type: (Solver, Dict[Player, Any], Optional[Lineup]) -> None
        pass


class NormalObjective(OptimizerRule):
    def __init__(self, optimizer, params):
        super(NormalObjective, self).__init__(optimizer, params)
        self.used_combinations = []  # type: List[Any]

    def apply(self, solver, players_dict):
        variables = []
        coefficients = []
        for player, variable in players_dict.items():
            variables.append(variable)
            coefficients.append(player.fppg)
        solver.set_objective(variables, coefficients)

    def apply_for_iteration(self, solver, players_dict, result):
        if not result:
            return
        self.used_combinations.append([players_dict[player] for player in result])
        total_players = self.optimizer.total_players
        for variables in self.used_combinations:
            solver.add_constraint(variables, None, SolverSign.LTE, total_players - 1)


class RandomObjective(OptimizerRule):
    def apply_for_iteration(self, solver, players_dict, result):
        variables = []
        coefficients = []
        for player, variable in players_dict.items():
            variables.append(variable)
            coefficients.append(player.fppg * (1 + (-1 if bool(getrandbits(1)) else 1) *
                                               uniform(*self.optimizer.get_deviation())))
        solver.set_objective(variables, coefficients)


class TotalPlayersRule(OptimizerRule):
    def apply(self, solver, players_dict):
        variables = players_dict.values()
        coefficients = [1] * len(variables)
        solver.add_constraint(variables, coefficients, SolverSign.EQ, self.optimizer.total_players)


class LineupBudgetRule(OptimizerRule):
    def apply(self, solver, players_dict):
        variables = []
        coefficients = []
        for player, variable in players_dict.items():
            variables.append(variable)
            coefficients.append(player.salary)
        solver.add_constraint(variables, coefficients, SolverSign.LTE, self.optimizer.budget)


class LockedPlayersRule(OptimizerRule):
    def __init__(self, optimizer, params):
        super(LockedPlayersRule, self).__init__(optimizer, params)
        self.used_players = defaultdict(int)  # type: DefaultDict[Player, int]
        self.total_lineups = params.get('n')
        self.remaining_iteration = params.get('n') + 1

    def apply_for_iteration(self, solver, players_dict, result):
        self.remaining_iteration -= 1
        for player, variable in players_dict.items():
            if not player.min_exposure:
                continue
            if ceil(player.min_exposure * self.total_lineups) >= \
                    self.remaining_iteration - self.used_players.get(player, 0):
                solver.add_constraint([variable], None, SolverSign.EQ, 1)
        if not result:
            # First iteration, locked players must have exposure > 0
            for player in self.optimizer.locked_players:
                solver.add_constraint([players_dict[player]], None, SolverSign.EQ, 1)
            return
        for player in result.lineup:
            self.used_players[player] += 1
        removed_players = []
        for player, used in self.used_players.items():
            max_exposure = player.max_exposure if player.max_exposure is not None else self.params.get('max_exposure')
            if max_exposure is not None and max_exposure <= used / self.total_lineups:
                removed_players.append(player)
                solver.add_constraint([players_dict[player]], None, SolverSign.EQ, 0)
        for player in self.optimizer.locked_players:
            if player not in removed_players:
                solver.add_constraint([players_dict[player]], None, SolverSign.EQ, 1)


class PositionsRule(OptimizerRule):
    def apply(self, solver, players_dict):
        extra_positions = self.optimizer.players_with_same_position
        positions = get_positions_for_optimizer(self.optimizer.settings.positions)
        unique_positions = self.optimizer.available_positions
        players_by_positions = {
            position: {variable for player, variable in players_dict.items()
                       if position in player.positions} for position in unique_positions
        }
        for position, places in positions.items():
            extra = 0
            if len(position) == 1:
                extra = extra_positions.get(position[0], 0)
            players_with_position = set()
            for pos in position:
                players_with_position.update(players_by_positions[pos])
            solver.add_constraint(players_with_position, None, SolverSign.GTE, places + extra)


class TeamMatesRule(OptimizerRule):
    def apply(self, solver, players_dict):
        for team, quantity in self.optimizer.players_from_one_team.items():
            players_from_same_team = [variable for player, variable in players_dict.items() if player.team == team]
            solver.add_constraint(players_from_same_team, None, SolverSign.EQ, quantity)


class MaxFromOneTeamRule(OptimizerRule):
    def apply(self, solver, players_dict):
        if not self.optimizer.max_from_one_team:
            return
        for team in self.optimizer.available_teams:
            players_from_team = [variable for player, variable in players_dict.items() if player.team == team]
            solver.add_constraint(players_from_team, None, SolverSign.LTE, self.optimizer.max_from_one_team)


class MinSalaryCapRule(OptimizerRule):
    def apply(self, solver, players_dict):
        variables = []
        coefficients = []
        for player, variable in players_dict.items():
            variables.append(variable)
            coefficients.append(player.salary)
        min_salary_cap = self.optimizer.min_salary_cap
        solver.add_constraint(variables, coefficients, SolverSign.GTE, min_salary_cap)


class FromSameTeamByPositionsRule(OptimizerRule):
    def apply(self, solver, players_dict):
        from_same_team = len(self.optimizer.positions_from_same_team)
        players_combinations = []
        # Create all possible combinations with specified position for each team
        for team in self.optimizer.available_teams:
            team_players = [player for player in players_dict.keys() if player.team == team]
            players_by_positions = []
            for position in self.optimizer.positions_from_same_team:
                players_by_positions.append([player for player in team_players if position in player.positions])
            for players_combination in product(*players_by_positions):
                # Remove combinations with duplicated players
                if len(set(players_combination)) != from_same_team:
                    continue
                players_combinations.append(tuple([player for player in players_combination]))
        combinations_variable = {combination: solver.add_variable(
            'combinations_%d' % i
        ) for i, combination in enumerate(players_combinations)}
        solver.add_constraint(combinations_variable.values(), None, SolverSign.GTE, 1)
        for combination in players_combinations:
            variables = [players_dict[player] for player in combination]
            solver.add_constraint(variables, None, SolverSign.GTE,
                                  from_same_team * combinations_variable[combination])


class RemoveInjuredRule(OptimizerRule):
    def apply(self, solver, players_dict):
        injured_players_variables = [variable for player, variable in players_dict.items() if player.is_injured]
        solver.add_constraint(injured_players_variables, None, SolverSign.EQ, 0)


class MaxRepeatingPlayersRule(OptimizerRule):
    def __init__(self, optimizer, params):
        super(MaxRepeatingPlayersRule, self).__init__(optimizer, params)
        self.exclude_combinations = []  # type: List[Any]

    def apply_for_iteration(self, solver, players_dict, result):
        max_repeating_players = self.optimizer.max_repeating_players
        if max_repeating_players is None or not result:
            return
        for players_combination in combinations(result.players, max_repeating_players + 1):
            self.exclude_combinations.append([players_dict[player] for player in players_combination])
        for players_combination in self.exclude_combinations:
            solver.add_constraint(players_combination, None, SolverSign.LTE, max_repeating_players)


class ProjectedOwnershipRule(OptimizerRule):
    def apply(self, solver, players_dict):
        min_variables = []
        min_coefficients = []
        max_variables = []
        max_coefficients = []
        min_projected_ownership = self.optimizer.min_projected_ownership
        max_projected_ownership = self.optimizer.max_projected_ownership
        for player, variable in players_dict.items():
            if player.projected_ownership is None:
                continue
            if min_projected_ownership:
                min_variables.append(variable)
                min_coefficients.append(player.projected_ownership - min_projected_ownership)
            if max_projected_ownership:
                max_variables.append(variable)
                max_coefficients.append(player.projected_ownership - max_projected_ownership)
        if min_variables:
            solver.add_constraint(min_variables, min_coefficients, SolverSign.GTE, 0)
        if max_variables:
            solver.add_constraint(max_variables, max_coefficients, SolverSign.LTE, 0)


class UniquePlayerRule(OptimizerRule):
    def apply(self, solver, players_dict):
        key_func = lambda t: t[0].full_name
        data = sorted(players_dict.items(), key=key_func)
        for player_id, group_iterator in groupby(data, key=key_func):
            group = list(group_iterator)
            if len(group) == 1:
                continue
            variables = [variable for player, variable in group]
            solver.add_constraint(variables, None, SolverSign.LTE, 1)


class LateSwapRule(OptimizerRule):
    def __init__(self, optimizer, params):
        super(LateSwapRule, self).__init__(optimizer, params)
        self.current_iteration = 0
        self.lineups = params.get('lineups')  # type: List[Lineup]

    def apply_for_iteration(self, solver, players_dict, result):
        current_lineup = self.lineups[self.current_iteration]
        unswappable_players = current_lineup.get_unswappable_players()
        remaining_positions = get_remaining_positions(self.optimizer.settings.positions, unswappable_players)
        # lock selected players
        for player in unswappable_players:
            solver.add_constraint([players_dict[player]], None, SolverSign.EQ, 1)
        # set remaining positions
        positions = get_positions_for_optimizer(remaining_positions)
        for position, places in positions.items():
            players_with_position = [variable for player, variable in players_dict.items()
                                     if list_intersection(position, player.positions) and
                                     player not in unswappable_players]
            solver.add_constraint(players_with_position, None, SolverSign.GTE, places)
        # Exclude players with active games
        for player, variable in players_dict.items():
            if player not in unswappable_players and player.is_game_started:
                solver.add_constraint([players_dict[player]], None, SolverSign.EQ, 0)
        self.current_iteration += 1


class TeamStacksRule(OptimizerRule):
    def apply(self, solver, players_dict):
        stacks = self.optimizer.team_stacks
        players_by_teams = {
            team: [player for player in players_dict.keys() if player.team == team]
            for team in self.optimizer.available_teams
        }

        stacks_dict = {}
        for i, stack in enumerate(sorted(stacks, reverse=True), start=1):
            stacks_dict[stack] = i

        for stack, total in stacks_dict.items():
            combinations_variables = []

            for team, players in players_by_teams.items():
                solver_variable = solver.add_variable('teams_stack_%d_%s' % (stack, team))
                combinations_variables.append(solver_variable)
                variables = [players_dict[player] for player in players]
                solver.add_constraint(variables, None, SolverSign.GTE,
                                      stack * solver_variable)
            solver.add_constraint(combinations_variables, None, SolverSign.GTE, total)


class RestrictPositionsForOpposingTeams(OptimizerRule):
    def apply(self, solver, players_dict):
        if not self.optimizer.opposing_teams_position_restriction:
            return
        for game in self.optimizer.games:
            first_team_players = {player: variable for player, variable in players_dict.items()
                                  if player.team == game.home_team}
            second_team_players = {player: variable for player, variable in players_dict.items()
                                   if player.team == game.away_team}
            for first_team_positions, second_team_positions in \
                    permutations(self.optimizer.opposing_teams_position_restriction, 2):
                first_team_variables = [variable for player, variable in first_team_players.items()
                                        if list_intersection(player.positions, first_team_positions)]
                second_team_variables = [variable for player, variable in second_team_players.items()
                                         if list_intersection(player.positions, second_team_positions)]
                for variables in product(first_team_variables, second_team_variables):
                    solver.add_constraint(variables, None, SolverSign.LTE, 1)



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry


class YahooSettings(BaseSettings):
    site = Site.YAHOO
    budget = 200
    max_from_one_team = 6


@SitesRegistry.register_settings
class YahooBasketballSettings(YahooSettings):
    sport = Sport.BASKETBALL
    positions = [
        LineupPosition('PG', ('PG', )),
        LineupPosition('SG', ('SG', )),
        LineupPosition('SF', ('SF', )),
        LineupPosition('PF', ('PF', )),
        LineupPosition('C', ('C', )),
        LineupPosition('G', ('PG', 'SG')),
        LineupPosition('F', ('SF', 'PF')),
        LineupPosition('UTIL', ('PG', 'SG', 'SF', 'PF', 'C'))
    ]


@SitesRegistry.register_settings
class YahooFootballSettings(YahooSettings):
    sport = Sport.FOOTBALL
    positions = [
        LineupPosition('QB', ('QB', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('WR', ('WR', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('RB', ('RB', )),
        LineupPosition('TE', ('TE', )),
        LineupPosition('FLEX', ('WR', 'RB', 'TE')),
        LineupPosition('DEF', ('DEF', ))
    ]


@SitesRegistry.register_settings
class YahooHockeySettings(YahooSettings):
    sport = Sport.HOCKEY
    positions = [
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('C', ('C', )),
        LineupPosition('C', ('C', )),
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('W', ('LW', 'RW')),
        LineupPosition('D', ('D', ))
    ]


@SitesRegistry.register_settings
class YahooBaseballSettings(YahooSettings):
    sport = Sport.BASEBALL
    positions = [
        LineupPosition('P', ('P',)),
        LineupPosition('P', ('P',)),
        LineupPosition('C', ('C',)),
        LineupPosition('1B', ('1B',)),
        LineupPosition('2B', ('2B',)),
        LineupPosition('3B', ('3B',)),
        LineupPosition('SS', ('SS',)),
        LineupPosition('OF', ('OF',)),
        LineupPosition('OF', ('OF',)),
        LineupPosition('OF', ('OF',)),
    ]


@SitesRegistry.register_settings
class YahooGolfSettings(YahooSettings):
    sport = Sport.GOLF
    positions = [
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
        LineupPosition('G', ('G', )),
    ]


@SitesRegistry.register_settings
class YahooSoccerSettings(YahooSettings):
    sport = Sport.SOCCER
    positions = [
        LineupPosition('GK', ('GK', )),
        LineupPosition('D', ('D', )),
        LineupPosition('D', ('D', )),
        LineupPosition('D', ('D', )),
        LineupPosition('M', ('M', )),
        LineupPosition('M', ('M', )),
        LineupPosition('M', ('M', )),
        LineupPosition('M', ('M', )),
        LineupPosition('F', ('F', )),
        LineupPosition('F', ('F', )),
        LineupPosition('UTIL', ('D', 'M', 'F')),
    ]



class Site:
    DRAFTKINGS = 'DRAFTKINGS'
    FANDUEL = 'FANDUEL'
    YAHOO = 'YAHOO'
    FANTASY_DRAFT = 'FANTASY_DRAFT'
    FANBALL = 'FANBALL'
    DRAFTKINGS_CAPTAIN_MODE = 'DRAFTKINGS_CAPTAIN_MODE'


class Sport:
    BASKETBALL = 'BASKETBALL'
    FOOTBALL = 'FOOTBALL'
    HOCKEY = 'HOCKEY'
    BASEBALL = 'BASEBALL'
    GOLF = 'GOLF'
    SOCCER = 'SOCCER'
    CANADIAN_FOOTBALL = 'CANADIAN_FOOTBALL'
    LEAGUE_OF_LEGENDS = 'LEAGUE_OF_LEGENDS'
    WNBA = 'WNBA'
    MMA = 'MMA'
